#ifndef COMPAT_H
#define COMPAT_H

#include "config.h"

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifndef HAVE_GETTIMEOFDAY
extern int gettimeofday(struct timeval * tv, struct timezone * tz);
# define HAVE_GETTIMEOFDAY
#endif

#ifndef HAVE_MEMCPY
/* this should use size_t, but some systems don't define it */
extern void * memcpy(void * dest, void const * src, unsigned long n);
# define HAVE_MEMCPY
#endif

#ifndef HAVE_MEMSET
/* this should use size_t, but some systems don't define it */
extern void * memset(void * dest, int c, unsigned long n);
# define HAVE_MEMSET
#endif

#ifndef HAVE_STRCASECMP
# ifdef HAVE_STRICMP
#  define strcasecmp(s1, s2) stricmp(s1, s2)
#  define HAVE_STRCASECMP
# else
extern int strcasecmp(char const * str1, char const * str2);
# define stricmp(s1, s2) strcasecmp(s1, s2)
# define HAVE_STRCASECMP
# define HAVE_STRICMP
# endif
#endif

#ifndef HAVE_STRDUP
extern char * strdup(char const * str);
# define HAVE_STRDUP
#endif

#ifndef HAVE_STRERROR
extern char const * strerror(int errornum);
# define HAVE_STRERROR
#endif

#ifndef HAVE_STRNCASECMP
extern int strncasecmp(char const * str1, char const * str2, unsigned int cnt);
# define HAVE_STRNCASECMP
#endif

#ifndef HAVE_STRSTR
extern char *strstr(const char *string, const char *pattern);
# define HAVE_STRSTR
#endif

#endif /* COMPAT_H */
