/* timeq.c - time-based event queue
 * Copyright 2000 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: timeq.c,v 1.7.2.1 2001/03/29 00:20:52 zoot Exp $
 */

#include "config.h"
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "common.h"
#include "heap.h"
#include "timeq.h"

heap_t timeq;

struct timeq_entry {
    timeq_func func;
    void *data;
};

void
timeq_init(void)
{
    timeq = heap_new(int_comparator);
}

time_t
timeq_next(void)
{
    void *time;
    heap_peek(timeq, &time, 0);
    return TIME_T_CAST(time);
}

void
timeq_add(time_t when, timeq_func func, void *data)
{
    struct timeq_entry *ent;
    void *w;
    ent = malloc(sizeof(struct timeq_entry));
    ent->func = func;
    ent->data = data;
    w = (void*)TIME_T_CAST(when);
    heap_insert(timeq, w, ent);
}

struct timeq_extra {
    time_t when;
    timeq_func func;
    void *data;
    int mask;
};

int
timeq_del_matching(void *key, void *data, void *extra)
{
    struct timeq_entry *a = data;
    struct timeq_extra *b = extra;
    int res = ((b->mask & TIMEQ_IGNORE_WHEN) || (TIME_T_CAST(key) == b->when))
	&& ((b->mask & TIMEQ_IGNORE_FUNC) || (a->func == b->func))
	&& ((b->mask & TIMEQ_IGNORE_DATA) || (a->data == b->data));
    if (res) free(data);
    return res;
}

void
timeq_del(time_t when, timeq_func func, void *data, int mask)
{
    struct timeq_extra extra;
    extra.when = when;
    extra.func = func;
    extra.data = data;
    extra.mask = mask;
    heap_remove_pred(timeq, timeq_del_matching, &extra);
}

void
timeq_run(time_t now)
{
    void *k, *d;
    struct timeq_entry *ent;
    while (1) {
	heap_peek(timeq, &k, &d);
	if (TIME_T_CAST(k) > now) break;
	ent = d;
	ent->func(ent->data);
	heap_pop(timeq);
        free(ent);
    }
}
